/*
 * Copyright (C) 2010 Novell, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 */

/*
 * Simple JSON to new-line delimited text convertor. Reads JSON text from STDIN,
 * producing an error message if the text is rejected.
 *
 * Usage example:
 *   % json < test/input/test1
 *
 * Will convert the following JSON in test/input/test1:
 *   { "glossary": {
 *       "title": "example glossary",
 *       "GlossDiv": {
 *         "title": "S",
 *         "GlossList": {
 *           "GlossEntry": {
 *             "ID": "SGML",
 *             "SortAs": "SGML",
 *             "GlossTerm": "Standard Generalized Markup Language",
 *             "Acronym": "SGML",
 *             "Abbrev": "ISO 8879:1986",
 *             "GlossDef": {
 *               "para": "A meta-markup language.",
 *               "GlossSeeAlso": ["GML", "XML"]
 *             },
 *             "GlossSee": "markup"
 *           }
 *         }
 *       }
 *     }
 *   }
 *
 * To:
 *   glossary/title : "example glossary"
 *   glossary/GlossDiv/title : "S"
 *   glossary/GlossDiv/GlossList/GlossEntry/ID : "SGML"
 *   glossary/GlossDiv/GlossList/GlossEntry/SortAs : "SGML"
 *   glossary/GlossDiv/GlossList/GlossEntry/GlossTerm : "Standard Generalized Markup Language"
 *   glossary/GlossDiv/GlossList/GlossEntry/Acronym : "SGML"
 *   glossary/GlossDiv/GlossList/GlossEntry/Abbrev : "ISO 8879:1986"
 *   glossary/GlossDiv/GlossList/GlossEntry/GlossDef/para : "A meta-markup language."
 *   glossary/GlossDiv/GlossList/GlossEntry/GlossDef/GlossSeeAlso/0 : "GML"
 *   glossary/GlossDiv/GlossList/GlossEntry/GlossDef/GlossSeeAlso/1 : "XML"
 *   glossary/GlossDiv/GlossList/GlossEntry/GlossSee : "markup"
 *
 * The desired locale can be set via an argument, eg:
 *   % json fr_FR.ISO-8859-1 < test/input/test1
 *
 * If the locale is not specified, the "C" locale is used.
 *
 * A simple test script is included. See run-tests.sh for details.
 *
 * Author: James Tan Y.J. <jatan@novell.com>
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>

#include "JSON_parser.h"

/* Stack function headers */
typedef struct stack
{
    char* data;
    struct stack *next;
} Stack;
void push(Stack **head, const char* str);
const char* pop(Stack **head);
void print_stack(Stack **head, int first);

/* Helper function headers */
static void handle_value(void);
static char* int2str(int i);
static int print(void* ctx, int type, const JSON_value* value);

/* Static variables */
static Stack *s_Keys;
static int s_ArrayIndex = 0,
           s_IsInArray = 0,
           s_IsFirstArrayElement = 0,
           s_IsArrayOfObjects = 0,
           s_ArrayObjectLevel = 0;

int main(int argc, char* argv[])
{
    s_Keys = malloc(sizeof(Stack));

    JSON_config config;
    init_JSON_config(&config);
    config.depth                  = 19;
    config.callback               = &print;
    config.allow_comments         = 0;
    config.handle_floats_manually = 1;

    if (argc >= 2) {
        if (!setlocale(LC_ALL, argv[1])) {
            fprintf(stderr, "Failed to set locale to '%s'\n", argv[1]);
        }
    }

    struct JSON_parser_struct* jc = NULL;
    jc = new_JSON_parser(&config);

    FILE* input = stdin;
    int count = 0, result = 0;
    for (; input ; ++count) {
        int next_char = fgetc(input);
        if (next_char <= 0) {
            break;
        }
        if (!JSON_parser_char(jc, next_char)) {
            fprintf(stderr, "JSON_parser_char: syntax error, byte %d\n", count);
            result = 1;
            goto done;
        }
    }
    if (!JSON_parser_done(jc)) {
        fprintf(stderr, "JSON_parser_end: syntax error\n");
        result = 1;
        goto done;
    }

done:
    delete_JSON_parser(jc);
    return result;
}

/* JSON parser callback function */
static int print(void* ctx, int type, const JSON_value* value)
{
    switch(type) {
    case JSON_T_OBJECT_BEGIN:
        if (s_IsInArray) {
            if (s_IsFirstArrayElement) {
                s_IsFirstArrayElement = 0; s_IsArrayOfObjects = 1;
            }
            if (s_IsArrayOfObjects) {
                s_ArrayObjectLevel++;
                if (s_ArrayObjectLevel == 1) {
                    push(&s_Keys, int2str(s_ArrayIndex++));
                }
            }
        }
        break;
    case JSON_T_OBJECT_END:
        pop(&s_Keys);
        if (s_IsInArray && s_IsArrayOfObjects) {
            s_ArrayObjectLevel--;
        }
        break;

    case JSON_T_ARRAY_BEGIN:
        s_IsInArray = 1; s_ArrayIndex = 0; s_IsFirstArrayElement = 1;
        s_ArrayObjectLevel = 0;
        break;
    case JSON_T_ARRAY_END:
        pop(&s_Keys);
        s_IsInArray = 0; s_ArrayIndex = 0;
        break;

    case JSON_T_KEY:
        push(&s_Keys, value->vu.str.value);
        break;

    case JSON_T_INTEGER:
        handle_value();
        printf(JSON_PARSER_INTEGER_SPRINTF_TOKEN"\n", value->vu.integer_value);
        break;
    case JSON_T_FLOAT:
        handle_value();
        printf("%f\n", value->vu.float_value);
        break;
    case JSON_T_NULL:
        print_stack(&s_Keys, 1);
        if (s_IsInArray && s_IsArrayOfObjects) {
            printf("/%d", s_ArrayIndex++);
        } else {
            pop(&s_Keys);
        }
        printf(" : null\n");
        break;
    case JSON_T_TRUE:
        handle_value();
        printf("true\n");
        break;
    case JSON_T_FALSE:
        handle_value();
        printf("false\n");
        break;
    case JSON_T_STRING:
        handle_value();
        printf("\"%s\"\n", value->vu.str.value);
        break;

    default:
        break;
    }

    return 1;
}

/* Helper functions */

static void handle_value(void)
{
    print_stack(&s_Keys, 1);
    if (s_IsInArray) {
        if (s_IsFirstArrayElement) {
            s_IsFirstArrayElement = 0; s_IsArrayOfObjects = 0;
        }
        if (s_IsArrayOfObjects) {
            pop(&s_Keys);
        } else {
            printf("/%d", s_ArrayIndex++);
        }
    } else {
        pop(&s_Keys);
    }
    printf(" : ");
}

static char* int2str(int i)
{
    char *str = malloc(5);
    sprintf(str, "%d", i);
    return str;
}

/* Stack functions */

void push(Stack **head, const char* str)
{
    Stack *node = malloc(sizeof(Stack));
    node->data  = malloc(strlen(str)+1);
    strcpy(node->data, str);
    node->next = *head;
    *head = node;
}

const char* pop(Stack **head)
{
    Stack *top = *head;
    const char* value = top->data;
    *head = top->next;
    free(top);
    return value;
}

void print_stack(Stack **head, int first)
{
    Stack *top = *head;
    if (top->next != NULL) {
        print_stack(&top->next, 0);
        printf("%s", top->data);
        if (!first) printf("/");
    }
}
